import numpy as np
import chaospy
import matplotlib.pyplot as plt
from scipy import stats
from mpl_toolkits.mplot3d import Axes3D

#script showing how does the Sobol sequence work and why it is bebtter than random sampling for our case

dim = 2
n = 100
np.zeros((2,2,2))
sobol_sequence2d = chaospy.create_sobol_samples(n, dim)

# Generate a random sample from the uniform distribution
random_sample = np.random.uniform(0, 1, size=dim*n)
x2 = random_sample[:n]
y2 = random_sample[n:]

plt.scatter(x2, y2, label='Random uniform', s=1)
plt.scatter(sobol_sequence2d[0, :], sobol_sequence2d[1, :], label='Sobol sequence', s =1)
plt.xlabel('X')
plt.ylabel('Y')
plt.title('2D Sobol sequence')
plt.legend()
plt.show()

#3D visualisation
dim = 3
n = 10
# Generate the Sobol sequence
sobol_sequence = chaospy.create_sobol_samples(n, dim)

x = sobol_sequence[0,:]
y = sobol_sequence[1,:]
z = sobol_sequence[2,:]

# Generate a random sample from the uniform distribution
random_sample = np.random.uniform(0, 1, size=dim*n)
x2 = random_sample[:n]
y2 = random_sample[n:2*n]
z2 = random_sample[2*n:]
# Create a figure and a 3D axes object
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')

# Create the scatter plot
ax.scatter(x, y, z, label='Sobol sequence')
ax.scatter(x2, y2, z2, label='random uniform sequence')

# Set labels and title
ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
ax.set_title('3D Sobol sequence')
ax.legend()
# Show the plot
plt.show()


disc = np.exp(- r * dt)
n_bins = int(np.sqrt(n))
n_bins = 100
ti = 50
n_assets = 2
s = [0.3,0.4]
spot_price = [100, 110]
r = 0.06
dt = 1/ti
# generating Sobol sequence
# uniform
sobol_sequence = chaospy.create_sobol_samples(n_bins, n_assets)
representative_paths = np.zeros((n_bins, ti + 1, n_assets))  #
# representative paths

for a in range(n_assets):
    a = 0
    sobol_inv_norm = stats.norm.ppf(sobol_sequence[a, :])
    sigma = s[a]
    sobol_increments = np.hstack((np.zeros((n_bins, 1)),
                                  (r - sigma ** 2 / 2) * dt + sigma * np.sqrt(dt) * np.reshape(sobol_inv_norm,
                                                                                               (n_bins, ti))))
    sobol_increments = np.hstack((np.zeros((n_bins, 1)),np.tile(np.reshape((r - sigma ** 2 / 2) * dt + sigma * np.sqrt(dt) * (sobol_inv_norm), (n_bins, 1)), ti)))
    representative_paths[:, :, a] = spot_price[a] * np.exp(np.cumsum(sobol_increments, axis=1))



np.tile(np.reshape((r - sigma ** 2 / 2) * dt + sigma * np.sqrt(dt) * (sobol_inv_norm), (n_bins, 1)),5)


from underlying import Underlying
from AMoption import Option
import numpy as np
from payoffs import *
n_bins = 100
T = 1
n_sims=10000

asset = Underlying(spot_price=100, r=0.06)
asset.calibrate_GBM(0.3,50)
option = Option(asset, payoff_func = lambda traj: payoff_creator_1d(traj, Call_Payoff, K=100, barrier=True, barrier_type='up-and-in',barrier_level=110), T = 1)
paths = asset.simulate(mode='GBM', size=10000, T=T).to_numpy()

if mode == "GBM":
    values_per_year = asset.values_per_year_GBM
elif mode == "Bootstrap":
    values_per_year = asset.values_per_year_Bootstrap
elif mode == "JD":
    values_per_year = asset.values_per_year_JD
else:
    raise Exception(
        'Longstaff-Schwartz method (LS) is supported only for GBM, JD and Boostrapping at this moment')

ti = round(values_per_year * T)
disc = np.exp(- asset.r * T / ti)

exercise = np.zeros((n_sims, ti+1))
# n of paths in a bin
n_paths = int(n_sims / n_bins)
sorted_steps = np.argsort(paths, axis=0)

all_payoffs = option.payoff_func(paths)
# all_payoffs = self.payoff_func(paths)

if not asset.ssp_calibrated:
    probs_matrix = np.zeros((n_bins, n_bins, ti))

for i in tqdm(range(ti + 1)):
    i=0
    cur_bins = sorted_steps[:, ti - i]
    hi = all_payoffs[cur_bins, ti - i]
    #hi = np.mean(hi.reshape(-1, n_paths), axis=1)  # payoffy w danym kroku

    if i == 0:
        V_i = hi

    else:
        V_i1 = V_i
        next_bins = sorted_steps[:, ti - i + 1]

        if asset.ssp_calibrated:
            probs = asset.probs_matrix[:, :, ti - i]
        else:
            probs = np.zeros((n_bins, n_bins))
            for j in range(n_bins):
                list1 = cur_bins[(j * n_paths): ((j + 1) * n_paths)]
                s = 0
                for k in range(n_bins):

                    list2 = next_bins[(k * n_paths): ((k + 1) * n_paths)]
                    # print(list2)
                    res = np.intersect1d(list1, list2)
                    # res = [x == y for x, y in zip(list1, list2)]
                    # print(len(res))
                    probs[j, k] = len(res) / n_paths
                    # print(res)
                    s += len(res)
                    if s == n_paths:
                        break
            probs_matrix[:, :, ti - i] = probs

        # print(probs)
        V_i = np.maximum(hi, probs @ V_i1 * disc)
        print(V_i.shape)
        print()
    exercise[:, ti - i] = V_i <= hi
if not asset.ssp_calibrated:
    asset.probs_matrix = probs_matrix
    asset.ssp_calibrated = True

Yt = hi * np.exp(- np.tile(np.arange(0, ti + 1), (n_sims, 1)) * (T/values_per_year) * asset.r)
exercise_time = np.argmax(exercise, axis=1)
res = np.mean(Yt[np.arange(len(Yt)), exercise_time])
self.upper_bound_price_ssp = np.mean(V_i)
self.ET_ssp = exercise.astype(int)


import numpy as np
import matplotlib.pyplot as plt
from AMoption import Option
from underlying import Underlying
from payoffs import *


n = 10000
T = 0.5
sigma1 = 0.4
sigma2 = 0.3
r = 0.06
N = 250
spot1 = 100
spot2 = 110
K1 = 100
K2 = 110

Aktywo = Underlying(spot_price=spot1, r=r)  # Aktywo = Underlying(100)
Aktywo.calibrate_GBM(sigma=sigma1, values_per_year_GBM=N)  # Aktywo.calibrate_GBM(0.3,0.07)

Aktywo2 = Underlying(spot_price=spot2, r=r)  # Aktywo = Underlying(100)
Aktywo2.calibrate_GBM(sigma=sigma2, values_per_year_GBM=N)  # Aktywo.calibrate_GBM(0.3,0.07)

Opcja = Option(underlyings=(Aktywo, Aktywo2), payoff_func=lambda traj: Double_Call_Payoff(traj, K1,K2), T=T)
Opcja2 = Option(underlyings=(Aktywo,Aktywo2), payoff_func=lambda traj: multiple_call_payoff(traj, [K1, K2]), T=T)
Opcja3 = Option(underlyings=Aktywo,  payoff_func=lambda traj: payoff_creator_1d(traj, Call_Payoff, path_dependent=True,
                                                                                pd_type='asian', K=K1), T=T)
Opcja4 = Option(underlyings=(Aktywo,Aktywo2), payoff_func=lambda traj: multiple_mixed_payoff(traj, [K1, K2],
                                                                                             ['call', 'put']), T=T)
#Options = [Opcja, Opcja2, Opcja3, Opcja4, Opcja5, Opcja6, Opcja7, Opcja8, Opcja9, Opcja10, Opcja11, Opcja12, Opcja13, Opcja14]
    #types = ['call', 'put', 'asian call', 'geometric asian', 'asian strike call', 'geometric asian strike call', 'lookback minimum call',
    #         'lookback max call', 'up-and-out call (140)', 'up and in call(140)', ]


def price_compare(option, n):
    print(option.payoff_func.__name__)
    print('State stace partitioning price:')
    print(option.ssp(n_sims=n))
    print('Upper estimate ssp:')
    print(option.upper_bound_price_ssp)
    print('LS price:')
    print(option.LS(n))
    print('European price:')
    print(option.european())


price_compare(Opcja, n)

price_compare(Opcja2, n)
Opcja2.ssp(n)
Opcja2.LS(n)
Opcja2.european()


asset1 = Underlying(spot_price=100, r=0.07)
asset1.calibrate_GBM(sigma=0.4, values_per_year_GBM=N)
asset1.calibrate_JD(sigma=0.4, jump_intensity=1 / 3)
asset2 = Underlying(spot_price=120, r=0.07)
asset2.calibrate_GBM(sigma=0.4, values_per_year_GBM=N)
asset2.calibrate_JD(sigma=0.2, jump_intensity=1 / 2)

option_3dim = Option((asset1, asset2, Aktywo), payoff_func=lambda traj: multiple_mixed_payoff(traj, 100,payoff_type=['put','put','put'], which ='min'), T=1)
option_4dim = Option((asset1, asset2, Aktywo, Aktywo2), payoff_func=lambda traj: multiple_put_payoff(traj, 100), T=1)



option_3dim.ssp(n)

option = Option(underlyings=asset1,
                payoff_func=lambda trajectory: payoff_creator_1d(trajectory, Call_Payoff, K=100, barrier=True,
                                                                 barrier_level=140), T=1)
print(option.ssp(10000, mode='JD'))
# 20.33463726670842 z permutacją 11s, 20.536608834016505
print(option.LS(10000, mode='JD'))


# 20.92430225200415 15s, 20.65867517809142

option = Option(underlyings=(asset1, asset2), payoff_func=lambda trajectories: double_max_put(trajectories, 100), T=1)
option.european(10000)
option.ssp(10000, 'JD')
option.LS(10000, 'JD')





# Calculate the median along each axis
#median_axis0 = np.quantile(arr, [0.25, 0.5, 0.75], axis=0)
#median_axis1 = np.quantile(arr, [0.25, 0.5, 0.75], axis=1)

# Create scatter plot
##x = np.arange(arr.shape[1])
#y = np.arange(arr.shape[0])

#fig, ax = plt.subplots()
#ax.scatter(x, median_axis0, label='Median along Axis 0')
#ax.scatter(median_axis1, y, label='Median along Axis 1')
#ax.legend()
#ax.set_xlabel('Index')
#x.set_ylabel('Value')
#ax.set_title('Quantiles in 2-Dimensional Array')
#plt.show()



import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt

def jump_diffusion(spot_price, T=1, mu=0.12, sigma=0.3, Lambda=0.25,
                   a=0.2, b=0.2, values_per_year=252, n_sim=100):
    '''
    Monte Carlo simulation [1] of Merton's Jump Diffusion Model [2].
    The model is specified through the stochastic differential equation (SDE):

                        dS(t)
                        ----- = mu*dt + sigma*dW(t) + dJ(t)
                        S(t-)

    with:

    mu, sigma: constants, the drift and volatility coefficients of the stock
               price process;
    W: a standard one-dimensional Brownian motion;
    J: a jump process, independent of W, with piecewise constant sample paths.
       It is defined as the sum of multiplicative jumps Y(j).

    Input
    ---------------------------------------------------------------------------
    S: float. The current asset price.
    T: int or float. The maturity of the option contract, i.e. the final
       monitoring date.
    mu, sigma: float. Respectively, the drift and volatility coefficients of
               the asset price process.
    Lambda: float. The intensity of the Poisson process in the jump diffusion
            model ('lambda' is a protected keyword in Python).
    a, b: float. Parameters required to calculate, respectively, the mean and
          variance of a standard lognormal distribution, log(x) ~ N(a, b**2).
          (see code).
    values_per_year: int. The number of monitoring dates, i.e. the time steps.
    n_sim: int. The number of Monte Carlo simulations (at least 10,000 required
          to generate stable results).

    '''


    # Calculate the length of the time step
    dt = T/values_per_year
    ti =values_per_year * T


    '''
    Generate an Nsim x (Nsteps+1) array of zeros to preallocate the simulated
    paths of the Monte Carlo simulation. Each row of the matrix represents a
    full, possible path for the stock, each column all values of the asset at
    a particular instant in time.
    '''
    simulated_paths = np.zeros([n_sim, ti+1])

    # Replace the first column of the array with the vector of initial price S
    simulated_paths[:,0] = spot_price

    '''
    To account for the multiple sources of uncertainty in the jump diffusion
    process, generate three arrays of random variables.

     - The first one is related to the standard Brownian motion, the component
       epsilon(0,1) in epsilon(0,1) * np.sqrt(dt);
     - The second and third ones model the jump, a compound Poisson process:
       the former (a Poisson process with intensity Lambda) causes the asset
       price to jump randomly (random timing); the latter (a Gaussian variable)
       defines both the direction (sign) and intensity (magnitude) of the jump.
    '''
    Z_1 = np.random.normal(size=[n_sim, ti])
    Z_2 = np.random.normal(size=[n_sim, ti])
    Poisson = np.random.poisson(Lambda*dt, [n_sim, ti])

    # Populate the matrix with Nsim randomly generated paths of length ti
    for i in range(ti):
        simulated_paths[:,i+1] = simulated_paths[:,i]*np.exp((mu
                               - sigma**2/2)*dt + sigma*np.sqrt(dt) \
                               * Z_1[:,i] + a*Poisson[:,i] \
                               + np.sqrt(b**2) * np.sqrt(Poisson[:,i]) \
                               * Z_2[:,i])
    return simulated_paths
# Single out array of simulated prices at maturity T
T=1
r=0.06
values_per_year=255
ti = T*values_per_year
sims = jump_diffusion(spot_price=100,T=T,mu=r,values_per_year=values_per_year)

# Choose palette, figure size, and define figure axes
sns.set(palette='viridis')
plt.figure(figsize=(10,8))
ax = plt.axes()

# Generate t, the time variable on the abscissae
t = np.linspace(0, T, values_per_year+1) * ti

# Plot the Monte Carlo simulated stock price paths
jump_diffusion = ax.plot(t, sims.transpose());

# Make drawn paths thinner by decreasing line width
plt.setp(jump_diffusion, linewidth=1);

# Set title (LaTeX notation) and x- and y- labels
ax.set(title="Monte Carlo simulated stock price paths in Merton's jump \
diffusion model\n$S_0$ = {}, $\mu$ = {}, $\sigma$ = {}, $a$ = {}, $b$ = {}, \
$\lambda$ = {}, $T$ = {}, Nsteps = {}, Nsim = {}"\
       .format(S, mu, sigma, a, b, Lambda, T, Nsteps, Nsim), \
       xlabel='Time (days)', ylabel='Stock price')

# Display figure in a Python environment
plt.show()
